/*
 * Configuration stuff (GTK) for the plugin.
 * Largely inspired from other plugins configuration.
 * A bit more cleaned up, though ? :)
 */

#include "fxt.h"
#include <gtk/gtk.h>

static struct FxtConfig backup_options;
static GtkWidget *dialog;

/*
 * restore original options
 */
static void
configure_cancel_cb(GtkWidget *w, gpointer data)
{
	plugin_cfg = backup_options;
	gtk_widget_destroy(dialog);
}


/*
 * keep & save options
 */
static void
configure_ok_cb(GtkWidget *w, gpointer data)
{ 
	fxt_cfg_save();
	gtk_widget_destroy(dialog);
}


/*
 * Create an horiz box with label and scale
 */
static GtkWidget *
labelled_scale_new (const char *title, GtkAdjustment *adj, gboolean isInteger)
{
	GtkWidget *hgroup;
	GtkWidget *label;
	GtkWidget *scale;

	hgroup = gtk_hbox_new(FALSE, 2);

	label = gtk_label_new(title);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(hgroup), label,
			   FALSE, TRUE, 2);

	scale = gtk_hscale_new(adj);
	gtk_box_pack_start(GTK_BOX(hgroup), scale,
			   TRUE, TRUE, 2);
	if (isInteger)
		gtk_scale_set_digits(GTK_SCALE(scale), 0);

	return hgroup;
}


/*
 * to modify an integer variable whenever an adjustment change.
 */
static void
intval_changed_cb (GtkAdjustment *adj, gpointer valptr)
{
	*(gint *)valptr = adj->value;
}

/*
 * to modify a float variable whenever an adjustment change.
 */
static void
floatval_changed_cb (GtkAdjustment *adj, gpointer valptr)
{
	*(gfloat *)valptr = adj->value;
}


/*
 * to modify a boolean variable whenever a button is toggled.
 */
static void
toggled_cb (GtkToggleButton *bt, gpointer valptr)
{
	*(gboolean *)valptr = bt->active;
}


/*
 * 
 */
static void
mode_toggled_cb (GtkToggleButton *bt, gpointer valptr)
{
	if (bt->active)
	{
		*(FXMode *)valptr = (FXMode)gtk_object_get_user_data(GTK_OBJECT(bt));
	}
}


/*
 * Create the configuration dialog. GTK isn't that readable, huh ? :)
 */
void 
fxt_configure(void)
{
	GtkWidget *frame, *vbox, *mode_box;
	GtkWidget *radio_bt, *ok_bt, *cancel_bt;
	GtkWidget *strength_group, *volume_group, *delay_group, *reverse_cm;
	GtkObject *strength_adj, *volume_adj, *delay_adj;
	GSList *mode_group;

	if (dialog)
		return;

	backup_options = plugin_cfg;
	dialog = gtk_dialog_new();
	gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
			   GTK_SIGNAL_FUNC(gtk_widget_destroyed), &dialog);
	gtk_window_set_title(GTK_WINDOW(dialog), "Configuration");
	gtk_container_border_width(GTK_CONTAINER(dialog), 5);

	frame = gtk_frame_new(PLUGIN_NAME);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), frame,
			   TRUE, TRUE, 5);

	vbox = gtk_vbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(frame), vbox);

	/* mode radio box */
	mode_box = gtk_hbox_new (FALSE, 10);
	gtk_box_pack_start(GTK_BOX(vbox), mode_box, TRUE, TRUE, 5);
	
	radio_bt = gtk_radio_button_new_with_label (NULL, "Surround");
	gtk_object_set_user_data(GTK_OBJECT(radio_bt), (gpointer)MODE_SURROUND);
	if (plugin_cfg.fxmode == MODE_SURROUND)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radio_bt), TRUE);
	gtk_signal_connect (GTK_OBJECT (radio_bt), "toggled",
			    GTK_SIGNAL_FUNC (mode_toggled_cb), &plugin_cfg.fxmode);
	gtk_box_pack_start(GTK_BOX(mode_box), radio_bt, TRUE, TRUE, 0);

	mode_group = gtk_radio_button_group (GTK_RADIO_BUTTON (radio_bt));

	radio_bt = gtk_radio_button_new_with_label (mode_group, "Echo");
	gtk_object_set_user_data(GTK_OBJECT(radio_bt), (gpointer)MODE_ECHO);
	if (plugin_cfg.fxmode == MODE_ECHO)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radio_bt), TRUE);
	gtk_signal_connect (GTK_OBJECT (radio_bt), "toggled",
			    GTK_SIGNAL_FUNC (mode_toggled_cb), &plugin_cfg.fxmode);
	gtk_box_pack_start(GTK_BOX(mode_box), radio_bt, TRUE, TRUE, 0);

	/* strength slider */
	strength_adj = gtk_adjustment_new(plugin_cfg.strength_pct,
					  0, 100, 1, 10, 0);
	gtk_signal_connect(GTK_OBJECT(strength_adj), "value_changed",
			   GTK_SIGNAL_FUNC(intval_changed_cb),
			   &plugin_cfg.strength_pct);

	strength_group = labelled_scale_new ("Strength",
					     GTK_ADJUSTMENT(strength_adj), TRUE);
	gtk_box_pack_start(GTK_BOX(vbox), strength_group, TRUE, TRUE, 5);
	
	/* delay slider */
	delay_adj = gtk_adjustment_new(plugin_cfg.delay_ms,
				       0, MAX_DELAY_MS, 0.1, 0.5, 0);
	gtk_signal_connect(GTK_OBJECT(delay_adj), "value_changed",
			   GTK_SIGNAL_FUNC(floatval_changed_cb),
			   &plugin_cfg.delay_ms);

	delay_group = labelled_scale_new ("Delay (ms)",
					  GTK_ADJUSTMENT(delay_adj), FALSE);
	gtk_box_pack_start(GTK_BOX(vbox), delay_group, TRUE, TRUE, 5);

	/* volume slider */
	volume_adj = gtk_adjustment_new(plugin_cfg.volume_pct,
					0, 100, 1, 10, 0);
	gtk_signal_connect(GTK_OBJECT(volume_adj), "value_changed",
			   GTK_SIGNAL_FUNC(intval_changed_cb),
			   &plugin_cfg.volume_pct);

	volume_group = labelled_scale_new ("Volume finetuning",
					   GTK_ADJUSTMENT(volume_adj), TRUE);
	gtk_box_pack_start(GTK_BOX(vbox), volume_group, TRUE, TRUE, 5);

	/* reverse checkmark */
	reverse_cm = gtk_check_button_new_with_label("Reverse stereo");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(reverse_cm),
				     plugin_cfg.reverse);
	gtk_signal_connect(GTK_OBJECT(reverse_cm), "toggled",
			   GTK_SIGNAL_FUNC(toggled_cb), &plugin_cfg.reverse);
	gtk_box_pack_start(GTK_BOX(vbox), reverse_cm, TRUE, TRUE, 5);

	/* dialog buttons */
	ok_bt = gtk_button_new_with_label("Ok");
	gtk_signal_connect(GTK_OBJECT(ok_bt), "clicked",
			   GTK_SIGNAL_FUNC(configure_ok_cb), NULL);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), ok_bt,
			   TRUE, TRUE, 0);

	cancel_bt = gtk_button_new_with_label("Cancel");
	gtk_signal_connect(GTK_OBJECT(cancel_bt), "clicked",
			   GTK_SIGNAL_FUNC(configure_cancel_cb), NULL);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), cancel_bt,
			   TRUE, TRUE, 0);
	
	gtk_widget_show_all(dialog);
	gtk_widget_grab_focus(ok_bt);
}
