/*
 * Attempt to make a surround effect. (idea from EaglePlayer2)
 * This is a so-called Digital Delayed Surround : you pick
 * a configurable amount of a channel, delay it for some millis,
 * and mix it to the other channel.
 *
 * Advantage : it gives you a larger room (just a feeling :), and
 * reduce pressure in headphones when hearing song with very
 * different audio channels (like a module). 
 *
 * Hu-ho, I did this plugin for x11amp, which hadn't such plugin.
 * Just before releeasing it, I've seen xmms, and the ton of fx plugins :)
 *
 * Added echo as a bonus, and renamed to "FX Toolbox"
 * "FXT, the sound of the galaxy" :)
 * First release : around July 28, 1999.
 * Slight API changes for xmms 0.9.5, Nov 23, 1999.
 */

#include "fxt.h"
#include "queue.h"

#include <xmms/configfile.h>

struct FxtConfig plugin_cfg;

static FifoQueue *fifo;


static void fxt_init(void);
static void fxt_cleanup(void);
static int fxt_modify_samples(gpointer *datap, int datasize,
			      AFormat fmt, int freq_hz, int nch);

static EffectPlugin fxt_ep =
{
	NULL,
	NULL,
	PLUGIN_NAME,

	fxt_init,
	fxt_cleanup,
	fxt_about,
	fxt_configure,
	fxt_modify_samples,
};
	

EffectPlugin *
get_eplugin_info(void)
{
	return &fxt_ep;
}

/* remove old api crap as soon as possible :) */
#define USE_OLD_CFG_API 1

#ifdef USE_OLD_CFG_API
static 
gboolean fxt_cfg_load()
{
	ConfigFile *cfg;
	gchar      *filename;

	filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);
	if((cfg = xmms_cfg_open_file(filename)) != NULL)
	{
		if (!xmms_cfg_read_int(cfg,"FXT", "mode", &plugin_cfg.fxmode))
		{
			xmms_cfg_free(cfg);
			return FALSE;
		}
		xmms_cfg_read_int(cfg,"FXT", "strength", &plugin_cfg.strength_pct);
		xmms_cfg_read_float(cfg, "FXT",	"delay", &plugin_cfg.delay_ms);
		xmms_cfg_read_int(cfg, "FXT", "volume", &plugin_cfg.volume_pct);
		xmms_cfg_read_boolean(cfg,"FXT", "reverse", &plugin_cfg.reverse);

		xmms_cfg_free(cfg);
		return TRUE;
	}
	else
		return FALSE;
}

void 
fxt_cfg_save()
{
	ConfigFile *cfg;
	gchar      *filename;

	filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);
	cfg = xmms_cfg_open_file(filename);
	if(!cfg)
		cfg = xmms_cfg_new();

	xmms_cfg_write_int(cfg, "FXT", "mode", plugin_cfg.fxmode);
	xmms_cfg_write_int(cfg, "FXT", "strength", plugin_cfg.strength_pct);
	xmms_cfg_write_float(cfg, "FXT", "delay", plugin_cfg.delay_ms);
	xmms_cfg_write_float(cfg, "FXT", "volume", plugin_cfg.volume_pct);
	xmms_cfg_write_boolean(cfg,"FXT", "reverse", plugin_cfg.reverse);
	xmms_cfg_write_file(cfg, filename);
	xmms_cfg_free(cfg);
	g_free(filename);
}

static void 
fxt_init(void)
{
	if (!fxt_cfg_load())
	{
	        plugin_cfg.fxmode = MODE_SURROUND;
		plugin_cfg.reverse = FALSE;
		plugin_cfg.strength_pct = 90;
		plugin_cfg.delay_ms = MAX_DELAY_MS / 4;
		plugin_cfg.volume_pct = 50;
	}
}

#else

static 
void fxt_cfg_load()
{
	ConfigFile *cfg;

	cfg = xmms_cfg_get_default_file(CFG_LOAD);

	xmms_cfg_read_int_def(cfg,"FXT", "mode", &plugin_cfg.fxmode, MODE_SURROUND);
	xmms_cfg_read_int_def(cfg,"FXT", "strength", &plugin_cfg.strength_pct, 90);
	xmms_cfg_read_float_def(cfg, "FXT", "delay", &plugin_cfg.delay_ms,
				MAX_DELAY_MS / (gfloat)4);
	xmms_cfg_read_int_def(cfg, "FXT", "volume", &plugin_cfg.volume_pct, 50);
	xmms_cfg_read_boolean_def(cfg,"FXT", "reverse", &plugin_cfg.reverse, FALSE);
	if (cfg)
		xmms_cfg_free(cfg);
}


void 
fxt_cfg_save()
{
	ConfigFile *cfg;

	cfg = xmms_cfg_get_default_file(CFG_SAVE);

	xmms_cfg_write_int(cfg, "FXT", "mode", plugin_cfg.fxmode);
	xmms_cfg_write_int(cfg, "FXT", "strength", plugin_cfg.strength_pct);
	xmms_cfg_write_float(cfg, "FXT", "delay", plugin_cfg.delay_ms);
	xmms_cfg_write_float(cfg, "FXT", "volume", plugin_cfg.volume_pct);
	xmms_cfg_write_boolean(cfg,"FXT", "reverse", plugin_cfg.reverse);
	xmms_cfg_write_default_file(cfg);
	xmms_cfg_free(cfg);
}

static void 
fxt_init(void)
{
	fxt_cfg_load();
}
#endif

static void 
fxt_cleanup(void)
{
    /* still crashy */
/*	surround_cfg_save();*/

	if (fifo)
		fifo_destroy(fifo);
}


/* datasize may vary from one call to another */
/*
 * code for stereo, 16 bit effect.
 * for each couple of left/right values:
 *   1. get current data
 *   2. store current data if fifo has some room
 *   3. retrieve delayed data. Skip to next samples if nothing to read
 *   4. add effect (echo or cross-echo)
 *   5  adjust output level by linear interpolation between
 *      "normalized" (lower) values and current values.
 *   6. clamp to sample size
 *   7. write modified sample, with optional channel reverse
 */
static void 
fxt_stereo16 (short *data, guint datasize,
	      guint byte_offset, FifoQueue *fifo)
{
	enum { TICK_SIZE = 4, NCHANS = 2, BYTES_PER_SAMPLE = 2,
	       MINVALUE = -32768, MAXVALUE = 32767};
	static char readbuf[TICK_SIZE];
	gint16 *sampleptr;
	gint16 *endptr = data + datasize / sizeof(short);
	int  left, right;
	gint16 left_echo, right_echo;
	static int cnt = 0;

	/* process each couple of left/right values */
	for (sampleptr = data; sampleptr < endptr; sampleptr += NCHANS)
	{
		/* get current data */
		left = sampleptr[0];
		right = sampleptr[1];

		/* store current data if fifo has some room */
		if (fifo_get_size(fifo) <= byte_offset)
			fifo_write(fifo, (char *)sampleptr, TICK_SIZE);
		/* retrieve delayed data. Skip to next samples if nothing to read */
		if (fifo_get_size(fifo) >= byte_offset)
			if (fifo_read(fifo, readbuf, TICK_SIZE) == -1)
				continue;
		
		right_echo = plugin_cfg.strength_pct
			* (*(short *)(readbuf + BYTES_PER_SAMPLE)) / 100;
		left_echo  = plugin_cfg.strength_pct
			* (*(short *)readbuf) / 100;
		
		/* add effect */
		if (plugin_cfg.fxmode == MODE_SURROUND)
		{
			left  += right_echo;
			right += left_echo;
		}
		else
		{
			left  += left_echo;
			right += right_echo;
		}

		/* adjust output level by linear interpolation between
		 * "normalized" (lower) values and current values.
		 */
		 if (plugin_cfg.volume_pct > 0)
		 {
		left  = ((100 * left / (100 + plugin_cfg.strength_pct)) *
			 (100 - plugin_cfg.volume_pct)
			 + left * plugin_cfg.volume_pct) / 100;
		right = ((100 * right / (100 + plugin_cfg.strength_pct)) *
			 (100 - plugin_cfg.volume_pct)
			 + right * plugin_cfg.volume_pct) / 100;
		}
		/* clamp to sample size */
		left  = CLAMP(left, MINVALUE, MAXVALUE);
		right = CLAMP(right, MINVALUE, MAXVALUE);

		if (plugin_cfg.reverse)
		{
			sampleptr[1] = left;
			sampleptr[0] = right;
		}
		else
		{
			sampleptr[0] = left;
			sampleptr[1] = right;
		}
	}
}


static int
fxt_modify_samples(gpointer *datap, int datasize, AFormat afmt, int freq_hz, int nch)
{
	int  byte_offset;
	int  tick_size;
	int  bits;

	if (!(afmt == FMT_S16_NE
	      || (afmt == FMT_S16_LE && G_BYTE_ORDER == G_LITTLE_ENDIAN)
	      || (afmt == FMT_S16_BE && G_BYTE_ORDER == G_BIG_ENDIAN))
	    || nch != 2)
		return datasize;
	if (afmt < FMT_U16_LE)
	    bits = 8;
	else
	    bits = 16;

	/* size of data for a sound tick during (1/freq) second */
	/* this is most often 4. If it should be more (!), increase MAX_TICK_SIZE */
	tick_size = (bits / 8) * nch;

	/* bytes to fifo for the delay */
	byte_offset = plugin_cfg.delay_ms * freq_hz * tick_size / 1000;

	/* fifo creation */
	if (!fifo ||
	    ((int)fifo_get_udata(fifo) != (tick_size * freq_hz)))
	{
		if (fifo)
			fifo_destroy(fifo);
		/* size must be a multiple of tick_size */
		fifo = fifo_new(tick_size * (freq_hz * MAX_DELAY_MS / 1000 + 1));
		fifo_set_udata(fifo, (gpointer)(tick_size * freq_hz));
	}

	if (bits == 16)
		fxt_stereo16(*((short **)datap), datasize, byte_offset, fifo);

	return datasize;
}


