#include <queue.h>

/*
 * FIFO byte queue implementation. For audio buffers.
 * No data is ever overwritten. Instead, writes will fail until
 * you free some room by reading.
 * (c) 2000 David Le Corfec, under GPL
 */

/*
 * It's currently not MT-safe, because I don't need it :)
 */

struct FifoQueue {
    guchar  *data;        /* memory */
    int      maxbytes;    /* memory size */

    guint    write_start; /* index to next write */ 
    guint    read_start;  /* index to next read */
    guint    used;        /* bytes in buffer (may never exceed maxbytes) */

    gpointer user_data;   /* for you to use */
};


/**
 * Create a queue
 *
 * bytes : buffer size in bytes.
 */
struct FifoQueue *
fifo_new (int bytes)
{
    struct FifoQueue *buffer;

    buffer = g_new0(struct FifoQueue, 1);

    buffer->data = g_malloc0(bytes);
    buffer->maxbytes = bytes;
    return buffer;
}


/**
 * destroy buffer.
 */
void
fifo_destroy (struct FifoQueue *fifo)
{
    if (fifo)
    {
	if (fifo->data)
	    g_free(fifo->data);
	fifo->data = NULL;
	g_free(fifo);
    }
}


/**
 * Reset.
 */
void
fifo_reset (struct FifoQueue *fifo)
{
    g_return_if_fail(fifo != NULL);

    fifo->used = fifo->read_start = fifo->write_start = 0;
}


/**
 * Get maxsize available for reading.
 */
int
fifo_get_size (struct FifoQueue *fifo)
{
    g_return_val_if_fail(fifo != NULL, -1);

    return fifo->used;
}


/**
 * get max buffer size.
 */
int
fifo_get_maxbytes (struct FifoQueue *fifo)
{
    g_return_val_if_fail(fifo != NULL, -1);

    return fifo->maxbytes;
}


/**
 * get user data.
 */
gpointer
fifo_get_udata (struct FifoQueue *fifo)
{
    g_return_val_if_fail(fifo != NULL, NULL);

    return fifo->user_data;
}


/**
 * set user data.
 */
void
fifo_set_udata (struct FifoQueue *fifo, gpointer data)
{
    g_return_if_fail(fifo != NULL);

    fifo->user_data = data;
}


/**
 * Read data from a queue.
 *
 * fifo            : queue to read from.
 * buf             : place to store read datas.
 * nbytes          : size to read (in bytes).
 *
 * return: number of bytes read, or -1 if buffer invalid.
 */
int
fifo_read (struct FifoQueue *fifo, char *buf, int nbytes)
{
    int start;

    g_return_val_if_fail(fifo != NULL, -1);
    g_return_val_if_fail(nbytes <= fifo->maxbytes, -1);

    start = fifo->read_start;

    /* trunc read if not enough to read */
    nbytes = MIN(nbytes, fifo->used);

    if (nbytes <= 0)
	return 0;
    
    if ((start + nbytes) <= fifo->maxbytes)
    {
	memcpy(buf, fifo->data + start, nbytes);
    }
    else
    {
	memcpy(buf, fifo->data + start, fifo->maxbytes - start);
	memcpy(buf + fifo->maxbytes - start, fifo->data,
	       nbytes - (fifo->maxbytes - start));
    }
    fifo->read_start = (start + nbytes) % fifo->maxbytes;
    fifo->used -= nbytes;

    return nbytes;
}


/**
 * Append datas to a queue.
 *
 * fifo   : queue to write into.
 * buf    : datas to write.
 * nbytes : number of bytes to write.
 *
 * return : number of bytes written, or -1 on error.
 */
int
fifo_write (struct FifoQueue *fifo, const char *buf, int nbytes)
{
    int start;

    g_return_val_if_fail(fifo != NULL, -1);
    g_return_val_if_fail(nbytes <= fifo->maxbytes, -1);

    start = fifo->write_start;

    /* trunc write so that its end doesn't overwrite beginning of read queue */
    nbytes = MIN(nbytes, fifo->maxbytes - fifo->used);

    if (nbytes <= 0)
	return 0;

    if ((start + nbytes) <= fifo->maxbytes)
    {
	memcpy(fifo->data + start, buf, nbytes);
    }
    else
    {
	memcpy(fifo->data + start, buf, fifo->maxbytes - start);
	memcpy(fifo->data, buf + fifo->maxbytes - start,
	       nbytes - (fifo->maxbytes - start));
    }
    fifo->write_start = (start + nbytes) % fifo->maxbytes;
    fifo->used += nbytes;

    return nbytes;
}


/*
 * debugging. dump n bytes starting from index a.
 */
void
fifo_dump (struct FifoQueue *fifo, int a, int n)
{
    int i;

    for (i = 0; i < n; i++)
    {
	g_print("%x.", fifo->data[i + a]);
    }
    g_print("\n");
}
